module Gargantext.Database.Action.Corpus where

import Data.Set.Internal qualified as Set
import Gargantext.API.Admin.Types (HasSettings)
import Gargantext.API.Errors.Types (BackendInternalError)
import Gargantext.Core.NodeStory.Types (HasNodeStoryEnv)
import Gargantext.Core.NLP (HasNLPServer)
import Gargantext.Core.Text.Corpus.Query qualified as Q
import Gargantext.Core.Text.List (buildNgramsLists)
import Gargantext.Core.Text.List.Group.WithStem (GroupParams (..))
import Gargantext.Core.Text.List.Social (FlowSocialListWith (..), FlowSocialListPriority (..))
import Gargantext.Core.Text.Ngrams (NgramsType (..))
import Gargantext.Core.Types.Individu (User (..))
import Gargantext.Core.Types.Main (ListType (..))
import Gargantext.Database.Action.Flow (getOrMkRootWithCorpus, reIndexWith)
import Gargantext.Database.Action.Flow.List (flowList_DbRepo)
import Gargantext.Database.Action.Search (searchInCorpus)
import Gargantext.Database.Action.User (getUserId)
import Gargantext.Database.Admin.Types.Hyperdata.Corpus (HyperdataCorpus)
import Gargantext.Database.Admin.Types.Node (CorpusId, nodeId2ContextId, NodeType (..))
import Gargantext.Database.Prelude (DBCmd')
import Gargantext.Database.Query.Facet.Types (facetDoc_id)
import Gargantext.Database.Query.Table.Node (insertDefaultNode, copyNode)
import Gargantext.Database.Query.Table.Node.Document.Add qualified as Document (add)
import Gargantext.Database.Query.Tree.Root (MkCorpusUser (..))
import Gargantext.Prelude


-- | Given a "parent" corpus and a query, search for all docs in the parent
--   that match the query, and create a corpus from those. The created corpus
--   is inserted in the tree as a child of the parent corpus.
--   Creation of subcorpus "Docs" and "Terms" nodes is handled. The terms can be
--   either copied from the parent corpus or recomputed based on the subcorpus docs.
makeSubcorpusFromQuery :: ( HasSettings     env
                          , HasNodeStoryEnv env
                          , HasNLPServer    env
                          )
  => User                                     -- ^ The corpus owner
  -> CorpusId                                 -- ^ ID of the parent corpus
  -> Q.Query                                  -- ^ The query to determine the subset of documents that will appear in the subcorpus
  -> Bool                                     -- ^ Whether to reuse parent term list (True) or compute a new one based only on the documents in the subcorpus (False)
  -> DBCmd' env BackendInternalError CorpusId -- ^ The child corpus ID
makeSubcorpusFromQuery user parentId query reuseParentList = do
  userId      <- getUserId user
  subcorpusId <- insertDefaultNode NodeCorpus parentId userId
  (_, _, masterCorpusId) <- getOrMkRootWithCorpus MkCorpusUserMaster (Nothing :: Maybe HyperdataCorpus)
  -- Get ahold of all documents that match the query
  facetDocs   <- searchInCorpus parentId False query Nothing Nothing Nothing
  -- Create a subcorpus node with all the documents
  _ <- Document.add subcorpusId $ nodeId2ContextId . facetDoc_id <$> facetDocs
  -- Create nodes for docs and terms as children of the subcorpus
  void $ insertDefaultNode NodeTexts subcorpusId userId
  listId <- insertDefaultNode NodeList subcorpusId userId
  -- Either simply copy parent terms... (TODO)
  if reuseParentList then void $ copyNode True _parentList subcorpusId
  -- ... or rebuild a term list from scratch
  else do
    ngrams <- buildNgramsLists
                user
                subcorpusId
                masterCorpusId
                (Just (FlowSocialListWithPriority MySelfFirst) :: Maybe FlowSocialListWith)
                GroupIdentity
    -- Save computed list
    _ <- flowList_DbRepo listId ngrams
    return ()
  -- Reindex 
  _ <- reIndexWith subcorpusId listId NgramsTerms (Set.singleton MapTerm)
  -- The following two lines (like the one just above) are copypasted from
  -- the definition of flowCorpusUser, but I'm not sure whether they should be included
  -- _ <- updateContextScore      userCorpusId listId
  -- _ <- updateNgramsOccurrences userCorpusId listId
  return subcorpusId

-- TODO
-- Permettre de (mettre les 2 en option) :
-- [X] 1a. relancer le buildNgramsLists (cf. module Flow)
-- [ ] 1b. ou copier la liste (WIP)
-- [ ] Vérifier l'étape de réindexation reIndexWith
-- [ ] 2. générer un graphe

